from __future__ import absolute_import, unicode_literals

import os

import django
import wagtail


PROJECT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
BASE_DIR = os.path.dirname(PROJECT_DIR)

DEBUG = True

ALLOWED_HOSTS = ['*']

SECRET_KEY = 'not needed'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'wagtailinventory.sqlite',
    },
}

if wagtail.VERSION >= (2, 0):
    WAGTAIL_APPS = (
        'wagtail.contrib.forms',
        'wagtail.contrib.settings',
        'wagtail.admin',
        'wagtail.core',
        'wagtail.documents',
        'wagtail.images',
        'wagtail.sites',
        'wagtail.users',
    )

    WAGTAIL_MIDDLEWARE = (
        'wagtail.core.middleware.SiteMiddleware',
    )

    WAGTAILADMIN_RICH_TEXT_EDITORS = {
        'default': {
            'WIDGET': 'wagtail.admin.rich_text.DraftailRichTextArea'
        },
        'custom': {
            'WIDGET': 'wagtail.tests.testapp.rich_text.CustomRichTextArea'
        },
    }
else:
    WAGTAIL_APPS = (
        'wagtail.contrib.settings',
        'wagtail.wagtailadmin',
        'wagtail.wagtailcore',
        'wagtail.wagtaildocs',
        'wagtail.wagtailforms',
        'wagtail.wagtailimages',
        'wagtail.wagtailsites',
        'wagtail.wagtailusers',
    )

    WAGTAIL_MIDDLEWARE = (
        'wagtail.wagtailcore.middleware.SiteMiddleware',
    )

    WAGTAILADMIN_RICH_TEXT_EDITORS = {
        'default': {
            'WIDGET': 'wagtail.wagtailadmin.rich_text.HalloRichTextArea',
        },
        'custom': {
            'WIDGET': 'wagtail.tests.testapp.rich_text.CustomRichTextArea'
        },
    }

if django.VERSION >= (1, 10):
    MIDDLEWARE = (
        'django.middleware.common.CommonMiddleware',
        'django.contrib.sessions.middleware.SessionMiddleware',
        'django.middleware.csrf.CsrfViewMiddleware',
        'django.contrib.auth.middleware.AuthenticationMiddleware',
        'django.contrib.messages.middleware.MessageMiddleware',
        'django.middleware.clickjacking.XFrameOptionsMiddleware',
    ) + WAGTAIL_MIDDLEWARE
else:
    MIDDLEWARE_CLASSES = (
        'django.middleware.common.CommonMiddleware',
        'django.contrib.sessions.middleware.SessionMiddleware',
        'django.middleware.csrf.CsrfViewMiddleware',
        'django.contrib.auth.middleware.AuthenticationMiddleware',
        'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
        'django.contrib.messages.middleware.MessageMiddleware',
        'django.middleware.clickjacking.XFrameOptionsMiddleware',
    ) + WAGTAIL_MIDDLEWARE

INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.messages',
    'django.contrib.sessions',
    'django.contrib.staticfiles',

    'taggit',
) + WAGTAIL_APPS + (
    'wagtailinventory',
    'wagtailinventory.tests.testapp',
)

STATIC_ROOT = os.path.join(BASE_DIR, 'static')
STATIC_URL = '/static/'

MEDIA_ROOT = os.path.join(BASE_DIR, 'media')
MEDIA_URL = '/media/'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
                'django.template.context_processors.request',
            ],
            'debug': True,
        },
    },
]

WAGTAIL_SITE_NAME = 'Test Site'

ROOT_URLCONF = 'wagtailinventory.tests.urls'



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion

try:
    from wagtail.core import blocks as core_blocks
    from wagtail.core import fields as core_fields  # pragma: no cover
except ImportError:  # pragma: no cover; fallback for Wagtail <2.0
    from wagtail.wagtailcore import blocks as core_blocks
    from wagtail.wagtailcore import fields as core_fields


class Migration(migrations.Migration):

    dependencies = [
        ('testapp', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='NestedStreamBlockPage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.Page')),
                ('content', core_fields.StreamField([('streamblock', core_blocks.StreamBlock([('text', core_blocks.CharBlock()), ('atom', core_blocks.StructBlock([('title', core_blocks.CharBlock())]))]))])),
            ],
            options={
                'abstract': False,
            },
            bases=('wagtailcore.page',),
        ),
    ]



from __future__ import absolute_import, unicode_literals

from django.db.models.fields import BLANK_CHOICE_DASH
from django.test import TestCase

try:
    from wagtail.core.models import Page, Site
except ImportError:  # pragma: no cover; fallback for Wagtail <2.0
    from wagtail.wagtailcore.models import Page, Site

from wagtailinventory.forms import PageBlockQueryForm
from wagtailinventory.models import PageBlock


class TestPageBlockQueryForm(TestCase):
    def test_no_page_blocks_in_database_form_has_blank_choice(self):
        form = PageBlockQueryForm()
        form_choices = form.fields['block'].choices

        self.assertEqual(len(form_choices), 1)
        self.assertEqual(form_choices, BLANK_CHOICE_DASH)

    def test_form_choices_alphabetized(self):
        root_page = Site.objects.get(is_default_site=True).root_page
        page = Page(title='Title', slug='title')
        root_page.add_child(instance=page)

        for block in (
            'blocks.mango',
            'blocks.apple',
            'blocks.banana',
        ):
            PageBlock.objects.create(page=page, block=block)

        form = PageBlockQueryForm()
        form_choices = form.fields['block'].choices

        self.assertEqual(len(form_choices), 4)
        self.assertEqual(form_choices[1][0], 'blocks.apple')
        self.assertEqual(form_choices[2][0], 'blocks.banana')
        self.assertEqual(form_choices[3][0], 'blocks.mango')



from __future__ import absolute_import, unicode_literals

from django.test import TestCase

try:
    from wagtail.core.models import Page
except ImportError:  # pragma: no cover; fallback for Wagtail <2.0
    from wagtail.wagtailcore.models import Page

from wagtailinventory.models import PageBlock


class TestPageBlock(TestCase):
    def test_page_str(self):
        page_block = PageBlock(
            page=Page(title='Title', slug='title'),
            block='path.to.block'
        )
        self.assertEqual(str(page_block), '<Title, path.to.block>')



try:
    from wagtail.core import blocks
except ImportError:  # pragma: no cover; fallback for Wagtail <2.0
    from wagtail.wagtailcore import blocks


class Atom(blocks.StructBlock):
    title = blocks.CharBlock()


class Molecule(blocks.StructBlock):
    title = blocks.CharBlock()
    atoms = blocks.ListBlock(Atom())


class Organism(blocks.StructBlock):
    molecules = blocks.ListBlock(Molecule())



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations

try:
    from wagtail.core import blocks as core_blocks
    from wagtail.core import fields as core_fields  # pragma: no cover
except ImportError:  # pragma: no cover; fallback for Wagtail <2.0
    from wagtail.wagtailcore import blocks as core_blocks
    from wagtail.wagtailcore import fields as core_fields


class Migration(migrations.Migration):

    dependencies = [
        ('testapp', '0002_nested_stream_block_page'),
    ]

    operations = [
        migrations.AlterField(
            model_name='singlestreamfieldpage',
            name='content',
            field=core_fields.StreamField([('atom', core_blocks.StructBlock([('title', core_blocks.CharBlock())])), ('molecule', core_blocks.StructBlock([('title', core_blocks.CharBlock()), ('atoms', core_blocks.ListBlock(core_blocks.StructBlock([('title', core_blocks.CharBlock())])))])), ('organism', core_blocks.StructBlock([('molecules', core_blocks.ListBlock(core_blocks.StructBlock([('title', core_blocks.CharBlock()), ('atoms', core_blocks.ListBlock(core_blocks.StructBlock([('title', core_blocks.CharBlock())])))])))]))], blank=True),
        ),
    ]



from itertools import chain

try:
    from wagtail.core.blocks import ListBlock, StreamBlock, StructBlock
    from wagtail.core.fields import StreamField  # pragma: no cover
except ImportError:  # pragma: no cover; fallback for Wagtail <2.0
    from wagtail.wagtailcore.blocks import ListBlock, StreamBlock, StructBlock
    from wagtail.wagtailcore.fields import StreamField

from wagtailinventory.models import PageBlock


def get_block_name(block):
    return block.__module__ + '.' + block.__class__.__name__


def get_page_blocks(page):
    blocks = []

    for field in page.specific._meta.fields:
        if not isinstance(field, StreamField):
            continue

        for stream_child in getattr(page.specific, field.name):
            blocks.extend(get_field_blocks(stream_child))

    return sorted(set(map(get_block_name, blocks)))


def get_field_blocks(value):
    block = getattr(value, 'block', None)
    blocks = [block] if block else []

    if isinstance(value, list):
        child_blocks = value
    if isinstance(block, StructBlock):
        if hasattr(value, 'bound_blocks'):
            child_blocks = value.bound_blocks.values()
        else:
            child_blocks = [value.value]
    elif isinstance(block, (ListBlock, StreamBlock)):
        child_blocks = value.value
    else:
        child_blocks = []

    blocks.extend(chain(*map(get_field_blocks, child_blocks)))

    return blocks


def get_page_inventory(page=None):
    inventory = PageBlock.objects.all()

    if page:
        inventory = inventory.filter(page=page)

    return inventory


def create_page_inventory(page):
    page_blocks = get_page_blocks(page)

    return [
        PageBlock.objects.get_or_create(page=page, block=block)[0]
        for block in page_blocks
    ]


def delete_page_inventory(page=None):
    get_page_inventory(page).delete()


def update_page_inventory(page):
    page_blocks = create_page_inventory(page)

    for page_block in get_page_inventory(page):
        if page_block not in page_blocks:
            page_block.delete()



from __future__ import absolute_import, unicode_literals

from django.http import HttpResponseBadRequest
from django.shortcuts import render
from django.views.generic import View
from wagtail.utils.pagination import paginate

try:
    from wagtail.core.models import Page
except ImportError:  # pragma: no cover; fallback for Wagtail <2.0
    from wagtail.wagtailcore.models import Page

from wagtailinventory.forms import PageBlockQueryFormSet


class SearchView(View):
    template_name = 'wagtailinventory/search.html'

    def get(self, request):
        if len(request.GET) > 1:
            formset = PageBlockQueryFormSet(request.GET)
            if not formset.is_valid():
                return HttpResponseBadRequest('invalid query')

            pages = formset.get_query()
        else:
            formset = PageBlockQueryFormSet()
            pages = Page.objects.all()

        paginator, pages = paginate(request, pages.order_by('title'))

        for page in pages:
            page.can_choose = True

        return render(request, self.template_name, {
            'formset': formset,
            'pages': pages,
        })



from django.test import TestCase

import wagtail

try:
    from wagtail.core.models import Page
except ImportError:  # pragma: no cover; fallback for Wagtail <2.0
    from wagtail.wagtailcore.models import Page

from wagtailinventory.helpers import get_page_blocks


if wagtail.VERSION[0] > 1:
    CORE_BLOCKS = 'wagtail.core.blocks'
else:  # pragma: no cover' fallback for Wagtail <2.0
    CORE_BLOCKS = 'wagtail.wagtailcore.blocks'


class TestGetPageBlocks(TestCase):
    fixtures = ['test_blocks.json']

    def test_page_with_no_streamfields_returns_empty_list(self):
        page = Page.objects.get(slug='no-streamfields-page')
        self.assertEqual(get_page_blocks(page), [])

    def test_empty_streamfield_returns_empty_list(self):
        page = Page.objects.get(slug='single-streamfield-page-no-content')
        self.assertEqual(get_page_blocks(page), [])

    def test_streamfield_with_single_block(self):
        page = Page.objects.get(slug='single-streamfield-page-content')
        self.assertEqual(
            get_page_blocks(page),
            [
                CORE_BLOCKS + '.field_block.CharBlock',
                'wagtailinventory.tests.testapp.blocks.Atom',
            ],
        )

    def test_multiple_streamfields(self):
        page = Page.objects.get(slug='multiple-streamfields-page')
        self.assertEqual(
            get_page_blocks(page),
            [
                CORE_BLOCKS + '.field_block.CharBlock',
                CORE_BLOCKS + '.list_block.ListBlock',
                'wagtailinventory.tests.testapp.blocks.Atom',
                'wagtailinventory.tests.testapp.blocks.Molecule',
                'wagtailinventory.tests.testapp.blocks.Organism',
            ]
        )

    def test_nested_streamblocks(self):
        page = Page.objects.get(slug='nested-streamblock-page')
        self.assertEqual(
            get_page_blocks(page),
            [
                CORE_BLOCKS + '.field_block.CharBlock',
                CORE_BLOCKS + '.stream_block.StreamBlock',
                'wagtailinventory.tests.testapp.blocks.Atom',
            ]
        )



from __future__ import absolute_import, unicode_literals

from django.db import models
from django.utils.encoding import python_2_unicode_compatible

try:
    from wagtail.core.models import Page
except ImportError:  # pragma: no cover; fallback for Wagtail <2.0
    from wagtail.wagtailcore.models import Page


@python_2_unicode_compatible
class PageBlock(models.Model):
    page = models.ForeignKey(Page, related_name='page_blocks',
                             on_delete=models.CASCADE)
    block = models.CharField(max_length=255, db_index=True)

    def __str__(self):
        return '<{}, {}>'.format(self.page, self.block)



from __future__ import absolute_import, unicode_literals

from django.conf.urls import include, url

try:
    from django.urls import reverse
except ImportError:  # pragma: no cover; fallback for Django <1.10
    from django.core.urlresolvers import reverse

try:
    from wagtail.admin.menu import MenuItem
    from wagtail.core import hooks  # pragma: no cover
except ImportError:  # pragma: no cover; fallback for Wagtail <2.0
    from wagtail.wagtailadmin.menu import MenuItem
    from wagtail.wagtailcore import hooks

from wagtailinventory import urls
from wagtailinventory.helpers import (
    create_page_inventory, delete_page_inventory, update_page_inventory
)


@hooks.register('after_create_page')
def do_after_page_create(request, page):
    create_page_inventory(page)


@hooks.register('after_edit_page')
def do_after_page_edit(request, page):
    update_page_inventory(page)


@hooks.register('after_delete_page')
def do_after_page_dete(request, page):
    delete_page_inventory(page)


@hooks.register('register_admin_urls')
def register_inventory_urls():
    return [
        url(r'^inventory/', include(urls, namespace='wagtailinventory')),
    ]


@hooks.register('register_settings_menu_item')
def register_inventory_menu_item():
    return MenuItem(
        'Block Inventory',
        reverse('wagtailinventory:search'),
        classnames='icon icon-placeholder',
        order=11000
    )



from django.db import models

try:
    from wagtail.admin.edit_handlers import FieldPanel, StreamFieldPanel
    from wagtail.core import blocks as wagtail_blocks  # pragma: no cover
    from wagtail.core.fields import StreamField  # pragma: no cover
    from wagtail.core.models import Page  # pragma: no cover
except ImportError:  # pragma: no cover; fallback for Wagtail <2.0
    from wagtail.wagtailadmin.edit_handlers import FieldPanel, StreamFieldPanel
    from wagtail.wagtailcore import blocks as wagtail_blocks
    from wagtail.wagtailcore.fields import StreamField
    from wagtail.wagtailcore.models import Page

from wagtailinventory.tests.testapp import blocks


class NoStreamFieldsPage(Page):
    content = models.TextField()

    content_panels = Page.content_panels + [
        FieldPanel('content'),
    ]


class SingleStreamFieldPage(Page):
    content = StreamField([
        ('atom', blocks.Atom()),
        ('molecule', blocks.Molecule()),
        ('organism', blocks.Organism()),
    ], blank=True)

    content_panels = Page.content_panels + [
        StreamFieldPanel('content'),
    ]


class MultipleStreamFieldsPage(Page):
    first = StreamField([
        ('atom', blocks.Atom()),
        ('molecule', blocks.Molecule()),
        ('organism', blocks.Organism()),
    ])
    second = StreamField([
        ('atom', blocks.Atom()),
        ('molecule', blocks.Molecule()),
        ('organism', blocks.Organism()),
    ])

    content_panels = Page.content_panels + [
        StreamFieldPanel('first'),
        StreamFieldPanel('second'),
    ]


class NestedStreamBlockPage(Page):
    content = StreamField([
        ('streamblock', wagtail_blocks.StreamBlock([
            ('text', wagtail_blocks.CharBlock()),
            ('atom', blocks.Atom()),
        ])),
    ])

    content_panels = Page.content_panels + [
        StreamFieldPanel('content'),
    ]



from __future__ import absolute_import, unicode_literals

from django import forms
from django.db.models.fields import BLANK_CHOICE_DASH
from django.forms import formset_factory

try:
    from wagtail.core.models import Page
except ImportError:  # pragma: no cover; fallback for Wagtail <2.0
    from wagtail.wagtailcore.models import Page

from wagtailinventory.models import PageBlock


class PageBlockQueryForm(forms.Form):
    INCLUDES_BLOCK = 'includes'
    EXCLUDES_BLOCK = 'excludes'

    block = forms.ChoiceField(
        choices=(),
        required=False,
        label='Block type',
        widget=forms.Select(attrs={'style': 'width: 50%'})
    )

    has = forms.ChoiceField(
        choices=((c, c) for c in (INCLUDES_BLOCK, EXCLUDES_BLOCK)),
        label=None,
        widget=forms.Select(attrs={'style': 'width: 100px'})
    )

    def __init__(self, *args, **kwargs):
        super(PageBlockQueryForm, self).__init__(*args, **kwargs)
        blocks = PageBlock.objects \
            .values_list('block', flat=True) \
            .distinct() \
            .order_by('block')

        block_choices = BLANK_CHOICE_DASH + [(b, b) for b in blocks]
        self.fields['block'].choices = block_choices


class BasePageBlockQueryFormSet(forms.BaseFormSet):
    def get_query(self):
        qs = Page.objects.all()

        for form in self.forms:
            if not form.is_valid():
                continue

            form_block = form.cleaned_data.get('block')

            if not form_block:
                continue

            condition = {'page_blocks__block': form_block}

            if form.cleaned_data['has'] == form.INCLUDES_BLOCK:
                qs = qs.filter(**condition)
            else:
                qs = qs.exclude(**condition)

        return qs


PageBlockQueryFormSet = formset_factory(
    PageBlockQueryForm,
    formset=BasePageBlockQueryFormSet,
    extra=3,
    max_num=3
)



from django.core.management import BaseCommand
from tqdm import tqdm

try:
    from wagtail.core.models import Page
except ImportError:  # pragma: no cover; fallback for Wagtail <2.0
    from wagtail.wagtailcore.models import Page

from wagtailinventory.helpers import (
    create_page_inventory, delete_page_inventory
)


class Command(BaseCommand):
    def handle(self, *args, **options):
        delete_page_inventory()

        for page in tqdm(Page.objects.all()):
            create_page_inventory(page)



from django.conf import settings
from django.conf.urls import include, url


try:
    from wagtail.admin import urls as wagtailadmin_urls
    from wagtail.core import urls as wagtailcore_urls  # pragma: no cover
    from wagtail.documents import urls as wagtaildocs_urls  # pragma: no cover
except ImportError:  # pragma: no cover; fallback for Wagtail <2.0
    from wagtail.wagtailadmin import urls as wagtailadmin_urls
    from wagtail.wagtailcore import urls as wagtailcore_urls
    from wagtail.wagtaildocs import urls as wagtaildocs_urls


urlpatterns = [
    url(r'^admin/', include(wagtailadmin_urls)),
    url(r'^documents/', include(wagtaildocs_urls)),
    url(r'', include(wagtailcore_urls)),
]


if settings.DEBUG:
    from django.conf.urls.static import static
    from django.contrib.staticfiles.urls import staticfiles_urlpatterns

    urlpatterns += staticfiles_urlpatterns()
    urlpatterns += static(
        settings.MEDIA_URL,
        document_root=settings.MEDIA_ROOT
    )



from setuptools import find_packages, setup


install_requires = [
    'Django>=1.8,<2.3',
    'tqdm==4.15.0',
    'wagtail>=1.8,<2.6',
]


setup_requires = [
    'setuptools-git-version==1.0.3',
]


testing_extras = [
    'coverage>=3.7.0',
    'flake8>=2.2.0',
    'mock>=1.0.0',
]


short_description = 'Lookup Wagtail pages by block content'


setup(
    name='wagtail-inventory',
    url='https://github.com/cfpb/wagtail-inventory',
    author='CFPB',
    author_email='tech@cfpb.gov',
    license='CCO',
    version_format='{tag}.dev{commitcount}+{gitsha}',
    include_package_data=True,
    packages=find_packages(),
    install_requires=install_requires,
    setup_requires=setup_requires,
    extras_require={
        'testing': testing_extras,
    },
    description=short_description,
    long_description=open('README.rst').read(),
    classifiers=[
        'Framework :: Django',
        'Framework :: Django :: 1.10',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'Framework :: Wagtail',
        'Framework :: Wagtail :: 1',
        'Framework :: Wagtail :: 2',
        'License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication',
        'License :: Public Domain',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ]
)
